/*
 *  Copyright (C) 2004 Steve Harris, Daniel Smith
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  $Id: mod_3store.c,v 1.23 2004/07/16 11:16:21 theno23 Exp $
 */

#include <httpd.h>
#include <ap_config.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_log.h>

#ifdef STANDARD20_MODULE_STUFF
#define VER2
#include <apr_strings.h>
#include <apr_compat.h>
#endif

#include "config.h"
#include "mod_3store.h"
#include "okbc.h"
#include "uri_escape.h"
#include "okbc_pp.h"
#include "handlers.h"
#include "hash_table.h"

#ifdef VER2
pthread_mutex_t strtok_m = PTHREAD_MUTEX_INITIALIZER;
module ts_module;
#define LOG(msg) ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_EMERG, \
			1, r->server, msg)
#else
module ts_module;
#define LOG(msg) ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_EMERG, \
			r->server, msg)
#endif

/* Fallback value for maximum RDQL query complexity */
#define DEF_MAX_COMP 1000

static void ts_parse_args(char *argvec[][], const char *args_in, int *count);

#ifdef VER2
static int ts_init(apr_pool_t* p, apr_pool_t* plog, apr_pool_t* ptemp,
		server_rec* s)
{
    okbc_init();
    okbc_local_connection();

    return OK;
}
#else
static void ts_init(server_rec *s, pool *p)
{
    okbc_init();
    okbc_local_connection();
}
#endif

static int ts_handle_okbc_req(request_rec *r)
{
    char *func_name;
    char *argvec[MAX_ARGS + 1][2];
    int argnum = 0, i, hash;
    okbc_node *ret;

#ifdef VER2
    if(strcmp(r->handler, "okbc-handler") != 0) {
	return DECLINED;
    }
#endif

    /* Default headers */
    r->content_type = "text/plain";
    r->status = OK;
    r->status_line = "200 OK";

    /* Make sure the arg string isn't (char *)NULL */
    if (!r->args) {
	r->args = "";
    }

    ts_parse_args(argvec, r->args, &argnum);

    func_name = r->path_info + 1;

    if (r->path_info == NULL || *func_name == '\0') {
	/* no function name provided */
	r->content_type = "text/html";
#ifndef VER2
	ap_send_http_header(r);
#endif
	ap_rprintf(r, "<html><head><title>OKBC Functions</title><body><h1>OKBC Functions</h1><ul>\n");
	for (i = 0; i < HASH_SIZE; i++) {
	    if (hash_table[i].name) {
		ap_rprintf(r, "<li><a href=\"http://www.ai.sri.com/~okbc/spec/okbc2/okbc2.html#%s\">%s</a> [<a href=\"%s\">call</a>]</li>\n",
			   hash_table[i].name, hash_table[i].name,
			   hash_table[i].name);
	    }
	}
	ap_rprintf(r, "</ul></body></html>\n");

	return OK;
    }

    /* Search for the requested function in the hashtable */
    hash = hash_func(func_name);
    while (hash_table[hash].name && strcmp(func_name, hash_table[hash].name)) {
	hash++;
    }

    if (!hash_table[hash].name) {
	/* We hit a gap in the hashtable before the function we want */
	r->content_type = "text/html";
	r->status_line = "404 No such function";
	r->status = HTTP_NOT_FOUND;
#ifndef VER2
	ap_send_http_header(r);
#endif
	ap_rprintf(r, "<html><head><title>%s</title><body><h1>%s: %s</h1><p>For more details see <A href=\"http://www.ai.sri.com/~okbc/spec/okbc2/okbc2.html\">the OKBC specification</a>.</p></body></html>\n",
		   r->status_line, r->status_line, func_name);

	return DONE;
    }

    /* Call the function ptr from the hashtable */
    ret = (hash_table[hash].func) (r, argvec);

    /* Free the memory used to escape URIs */
    for (i=0; i<argnum; i++) {
	    free(argvec[i][0]);
	    free(argvec[i][1]);
    }

    if (r->status == OK) {
	char *fmt = has_arg(argvec, "fmt");

	/* Chose the right prettyprinter */
	if (!fmt || !strcmp(fmt, "xml")) {
	    r->content_type = "text/xml";
#ifndef VER2
	    ap_send_http_header(r);
#endif
	    node_pp_xml(r, ret);
	} else {
#ifndef VER2
	    ap_send_http_header(r);
#endif
	    node_pp_sexp(r, ret);
	}
    } else {
	/* Dump the error message */
	r->content_type = "text/html";
#ifndef VER2
	ap_send_http_header(r);
#endif
	ap_rprintf(r, "<html><head><title>%s</title><body><h1>%s</h1><p>For more details see <A href=\"http://www.ai.sri.com/~okbc/spec/okbc2/okbc2.html#%s\">the OKBC specification</a>.</p></body></html>\n",
		   r->status_line, r->status_line, func_name);

	return DONE;
    }

    return r->status;
}

static int ts_handle_rdql_req(request_rec *r)
{
    char *argvec[MAX_ARGS + 1][2];
    char *query;
    int argnum;
    unsigned int i;
    char *show_sql;
    char *max_comp_str;
    char *xslt;
    int max_comp;
    rs_rdql_result *res;
    rs_rdql_column *row;
    ts_path_conf *cfg = (ts_path_conf *)
		ap_get_module_config(r->per_dir_config, &ts_module);

#ifdef VER2
    if(strcmp(r->handler, "rdql-handler") != 0) {
	return DECLINED;
    }
#endif

    /* Default headers */
    r->content_type = "text/xml";
    r->status = OK;
    r->status_line = "200 OK";

    ts_parse_args(argvec, r->args, &argnum);

    query = has_arg(argvec, "query");

    if (!query) {
        r->content_type = "text/html";
#ifndef VER2
        ap_send_http_header(r);
#endif
        ap_rprintf(r, "<html><head><title>3store: RDQL</title><body><h1>3store: RDQL Query</h1><form action=\".\"><textarea name=\"query\" cols=\"80\" rows=\"24\"></textarea><br><input type=\"submit\"></form></body></html>\n");

        return r->status;
    }

#ifndef VER2
    ap_send_http_header(r);
#endif

    show_sql = has_arg(argvec, "show_sql");

    max_comp_str = has_arg(argvec, "max_comp");
    if (max_comp_str) {
	max_comp = atoi(max_comp_str);
    } else {
	max_comp = DEF_MAX_COMP;
    }

    xslt = has_arg(argvec, "xslt");

    if (!cfg->db) {
	cfg->db = rs_connect_remote(cfg->remote, cfg->database, cfg->username,
				    cfg->password, NULL);
	if (!cfg->db) {
	    LOG("3store: configuration error, unable to connect to KB.");
	    r->status = HTTP_INTERNAL_SERVER_ERROR;
	    r->status_line = "500 3store: Unable to connect to KB";

	    return r->status;
	}
    }
    ap_rprintf(r, "<?xml version='1.0' encoding='UTF-8'?>\n");
    res = rs_rdql_query(cfg->db, query, max_comp);
    if (res->error_str) {
	ap_rprintf(r, "<error>%s</error>\n", res->error_str);
	return r->status;
    }
    if (xslt && *xslt) {
	ap_rprintf(r, "<?xml-stylesheet type=\"text/xsl\" href=\"%s\" ?>\n",
		   xslt);
    }
    if (show_sql) {
	ap_rprintf(r, "<!-- %s -->\n", res->sql_query);
    }
    ap_rprintf(r, "<table>\n");

    while ((row = rs_rdql_next_row(res))) {
        ap_rprintf(r, "<row>\n");
        for (i=0; i<res->columns; i++) {
	    if (row[i].value) {
	        if (strlen(row[i].value) > strcspn(row[i].value, "<>&")) {
		    char *tmp = xml_escape(row[i].value);
		    ap_rprintf(r, "<column name=\"%s\" type=\"%s\">%s</column>\n", row[i].variable, row[i].type == ObjURI ? "uri" : "literal", tmp);
		    free(tmp);
		} else {
		    ap_rprintf(r, "<column name=\"%s\" type=\"%s\">%s</column>\n", row[i].variable, row[i].type == ObjURI ? "uri" : "literal", row[i].value);
		}
	    }
        }
        ap_rprintf(r, "</row>\n");
    }
    rs_rdql_free_result(res);
    ap_rprintf(r, "</table>\n");

    return r->status;
}

static void ts_parse_args(char *argvec[MAX_ARGS + 1][2], const char *args_in,
			  int *count)
{
    char *tmp, *arg, *args;
    char divider;
    unsigned int i;

    if (!args_in || !(*args_in)) {
        *count = 0;
	argvec[0][0] = NULL;
	argvec[0][1] = NULL;

	return;
    }
    args = strdup(args_in);

    *count = 0;
    tmp = args;
#ifdef VER2
    pthread_mutex_lock(&strtok_m);
#endif
    while ((arg = strtok(tmp, "&?")) && *count < MAX_ARGS) {
	for (i = 0; arg[i] && arg[i] != '='; i++) { }
	divider = arg[i];
	arg[i] = '\0';
	argvec[*count][0] = uri_unescape(arg);
	if (divider) {
	    argvec[*count][1] = uri_unescape(arg + i + 1);
	} else {
	    argvec[*count][1] = NULL;
	}
	(*count)++;
	tmp = NULL;
    }
#ifdef VER2
    pthread_mutex_unlock(&strtok_m);
#endif
    argvec[*count][0] = NULL;
    argvec[*count][1] = NULL;

    free(args);
}

#ifdef VER2
static void *create_ts_config(apr_pool_t *p, server_rec *s)
#else
static void *create_ts_config(pool *p, server_rec *s)
#endif
{
    ts_server_conf *c = (ts_server_conf *)ap_palloc(p, sizeof(ts_server_conf));
    /* nothing there at the moment */

    return c;
}

#ifdef VER2
static void *create_ts_dir_config(apr_pool_t *p, char *dir)
#else
static void *create_ts_dir_config(pool *p, char *dir)
#endif
{
    ts_path_conf *c = (ts_path_conf *)ap_palloc(p, sizeof(ts_path_conf));
    c->username = RS_DB_USER;
    c->password = RS_DB_PASSWORD;
    c->database = RS_DB_DATABASE;
    c->remote = NULL;
    c->db = NULL;

    return (void *)c;
}

/* Apache internal stuff */

static const char *set_username(cmd_parms *cmd, void *mconfig, const char *user)
{
    ts_path_conf *cfg = (ts_path_conf *)mconfig;

    cfg->username = ap_pstrdup(cmd->pool, user);

    return NULL;
}

static const char *set_password(cmd_parms *cmd, void *mconfig, const char *user)
{
    ts_path_conf *cfg = (ts_path_conf *)mconfig;

    cfg->password = ap_pstrdup(cmd->pool, user);

    return NULL;
}

static const char *set_remote(cmd_parms *cmd, void *mconfig, const char *user)
{
    ts_path_conf *cfg = (ts_path_conf *)mconfig;

    cfg->remote = ap_pstrdup(cmd->pool, user);

    return NULL;
}

static const char *set_database(cmd_parms *cmd, void *mconfig, const char *user)
{
    ts_path_conf *cfg = (ts_path_conf *)mconfig;

    cfg->database = ap_pstrdup(cmd->pool, user);

    return NULL;
}

#ifndef VER2
#define AP_INIT_TAKE1(directive, func, mconfig, where, help) \
    { directive, func, mconfig, where, TAKE1, help }
#endif

static const command_rec ts_commands[] =
{
    AP_INIT_TAKE1("3SUsername", set_username, NULL, ACCESS_CONF,
		  "the username to used in 3store connections"),
    AP_INIT_TAKE1("3SPassword", set_password, NULL, ACCESS_CONF,
		  "the password to used in 3store connections"),
    AP_INIT_TAKE1("3SRemote", set_remote, NULL, ACCESS_CONF,
		  "the remote host to used for 3store connections"),
    AP_INIT_TAKE1("3SDatabase", set_database, NULL, ACCESS_CONF,
		  "the database to used in 3store connections"),
    {NULL}
};

#ifndef VER2
static const handler_rec ts_handlers[] = {
    {"okbc-handler", ts_handle_okbc_req},
    {"rdql-handler", ts_handle_rdql_req},
    {NULL}
};

module MODULE_VAR_EXPORT ts_module = {
    STANDARD_MODULE_STUFF,
    ts_init,			/* initializer */
    create_ts_dir_config,	/* dir config creator */
    NULL,			/* dir merger --- default is to override */
    create_ts_config,		/* server config */
    NULL,			/* merge server config */
    ts_commands,		/* command table */
    ts_handlers,		/* response handlers */
    NULL,			/* filename translation */
    NULL,			/* check_user_id */
    NULL,			/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
};
#else
static void ts_register_hooks(apr_pool_t* p)
{
    ap_hook_post_config (ts_init,            NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_handler     (ts_handle_okbc_req, NULL, NULL, APR_HOOK_FIRST );
    ap_hook_handler     (ts_handle_rdql_req, NULL, NULL, APR_HOOK_FIRST );
}

module AP_MODULE_DECLARE_DATA ts_module =
{
    STANDARD20_MODULE_STUFF,
    create_ts_dir_config,   /* create per-directory config structures */
    NULL,		    /* merge per-directory config structures  */
    create_ts_config,	    /* create per-server config structures    */
    NULL,		    /* merge per-server config structures     */
    ts_commands,	    /* command handlers */
    ts_register_hooks	    /* register hooks   */
};
#endif

/* vi:set ts=8 sts=4 sw=4: */
